package metrics

import (
	"go.opentelemetry.io/otel"
	otelmetric "go.opentelemetry.io/otel/metric"
)

type CacheMetrics struct {
	HitsTotal        otelmetric.Int64Counter
	MissesTotal      otelmetric.Int64Counter
	InvalidatesTotal otelmetric.Int64Counter
}

func GetCacheMetrics() (*CacheMetrics, error) {
	var (
		meter        = otel.Meter("")
		cacheMetrics = new(CacheMetrics)
		err          error
	)

	if cacheMetrics.HitsTotal, err = meter.Int64Counter("cache_hits",
		otelmetric.WithDescription("Количество найденных в кэше значений"),
	); err != nil {
		return nil, err
	}

	if cacheMetrics.MissesTotal, err = meter.Int64Counter("cache_misses",
		otelmetric.WithDescription("Количество не найденных в кэше значений"),
	); err != nil {
		return nil, err
	}

	if cacheMetrics.InvalidatesTotal, err = meter.Int64Counter("cache_invalidates",
		otelmetric.WithDescription("Количество инвалидаций кэша"),
	); err != nil {
		return nil, err
	}

	return cacheMetrics, nil
}
