package metrics

import (
	"go.opentelemetry.io/otel"
	otelmetric "go.opentelemetry.io/otel/metric"
)

type RequestMetrics struct {
	Total                otelmetric.Int64Counter
	FailedTotal          otelmetric.Int64Counter
	DurationMilliseconds otelmetric.Int64Histogram
}

// GetRequestMetrics возвращает метрики для подсчета количества удачных/неудачных запросов, а так же длительности ответов.
//
// Для RequestMetrics.DurationMilliseconds значения buckets по умолчанию равно []float64{0, 5, 10, 25, 50, 75, 100, 250, 500, 750, 1000, 2500, 5000, 7500, 10000}
func GetRequestMetrics() (*RequestMetrics, error) {
	var (
		meter          = otel.Meter("")
		requestMetrics = new(RequestMetrics)
		err            error
	)

	if requestMetrics.Total, err = meter.Int64Counter("requests",
		otelmetric.WithDescription("Количество запросов"),
	); err != nil {
		return nil, err
	}

	if requestMetrics.FailedTotal, err = meter.Int64Counter("requests_failed",
		otelmetric.WithDescription("Количество запросов, вернувших ошибку"),
	); err != nil {
		return nil, err
	}

	if requestMetrics.DurationMilliseconds, err = meter.Int64Histogram("request_duration",
		otelmetric.WithDescription("Длительность обработки запроса"),
		otelmetric.WithUnit("ms"),
	); err != nil {
		return nil, err
	}

	return requestMetrics, nil
}
